/* $Id$ */
/* David Leonard, 2001. Public domain. */

#include <assert.h>
#include <util.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <err.h>
#include <string.h>
#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sys/queue.h>

char version[] = "1.1";

/*
 * Forward data between two files (tun devices usually), but delaying the
 * data by the specified number of seconds
 */

struct packet {
	struct timespec when;	/* when to dequeue this packet */
	size_t	len;
	void	*data;
	int	fd;		/* file descriptor to send on on */
	TAILQ_ENTRY(packet) qe;
};

TAILQ_HEAD(queue, packet);

/* Allocate new queueable packet */
struct packet *
new_packet(len)
	size_t len;
{
	struct packet *p;

	p = (struct packet *)malloc(sizeof (struct packet));
	if (p == NULL)
		errx(1, "malloc");
	p->data = malloc(len);
	if (p->data == NULL)
		errx(1, "malloc");
	p->len = len;
	return p;
}

/* Free queueable packet */
void
free_packet(p)
	struct packet *p;
{
	free(p->data);
	free(p);
}

int
main(argc, argv)
	int argc;
	char *argv[];
{
	struct timespec ts, now, *tsp;
	struct timespec delay = { 30, 0 };
	struct queue q;
	int kq;
	int t1, t2;
	int nevts;
	struct kevent ev;
	struct packet *top, *p;
	char *path;
	ssize_t len;
	int error, ch;
	extern int optind;
	extern char *optarg;
	char *progname;

	/* process command line arguments */
	progname = argv[0];
	error = 0;
	while ((ch = getopt(argc, argv, "d:")) != -1) 
		switch (ch) {
		case 'd':
			delay.tv_sec = atoi(optarg);
			break;
		default:
			error = 1;
			break;
		}
	argc -= optind;
	argv += optind;
	if (argc != 2)
		error = 1;
	if (error) {
		fprintf(stderr, "usage: %s [-d delay] tun0 tun1\n", progname);
		exit(1);
	}

	/* open the two given devices */
	t1 = opendev(argv[0], O_RDWR, 0, &path);
	if (t1 == -1)
		err(1, "%s", path);
	t2 = opendev(argv[1], O_RDWR, 0, &path);
	if (t2 == -1)
		err(1, "%s", path);

	/* allocate a kernel event queue */
	kq = kqueue();
	if (kq == -1)
		err(1, "kqueue");
	EV_SET(&ev, t1, EVFILT_READ, EV_ADD, 0, 0, (void *)t2);
	if (kevent(kq, &ev, 1, NULL, 0, NULL) == -1)
		err(1, "kevent");
	EV_SET(&ev, t2, EVFILT_READ, EV_ADD, 0, 0, (void *)t1);
	if (kevent(kq, &ev, 1, NULL, 0, NULL) == -1)
		err(1, "kevent");

	/* initialise our timed buffer queue */
	TAILQ_INIT(&q);

	if (clock_gettime(CLOCK_REALTIME, &now) == -1)
		err(1, "clock_gettime");
	top = NULL;
	for (;;) {
		/* determine if we need to timeout */
		if (!top) {
			tsp = NULL;
		} else {
			assert(timespeccmp(&now, &top->when, <));
			timespecsub(&top->when, &now, &ts);
			tsp = &ts;
		}

		/* wait for an event or timeout */
		nevts = kevent(kq, NULL, 0, &ev, 1, tsp);
		if (nevts == -1)
			err(1, "kevent");

		if (clock_gettime(CLOCK_REALTIME, &now) == -1)
			err(1, "clock_gettime");
		if (nevts == 1) {
			if (ev.flags & EV_ERROR)
				errx(1, "kevent: %s", strerror((int)ev.data));

			/* new data arrived - read & add to queue */
			p = new_packet(ev.data);
			p->fd = (int)ev.udata;
			len = read(ev.ident, p->data, ev.data);
			if (len == -1)
				err(1, "read");
			assert(len == ev.data);
			timespecadd(&now, &delay, &p->when);
			TAILQ_INSERT_TAIL(&q, p, qe);
			top = TAILQ_FIRST(&q);
		}

		/* get rid of any expired things on the queue */
		while (top && timespeccmp(&top->when, &now, <=)) {
			len = write(top->fd, top->data, top->len);
			if (len == -1)
				err(1, "write");
			assert (len == top->len);
			TAILQ_REMOVE(&q, top, qe);
			free_packet(top);
			top = TAILQ_FIRST(&q);
		}
	}
}
